import json
import string    
import random

from treelib import Node, Tree


class Treeshow:
    def __init__(self, tree: Tree):
        self.tree = tree

    def url_list_to_json_tree(self, urls: list[str]):
        paths: list[list[str]] = []
        for item in urls:
            split = item.split('/')
            paths.append(split[2:-1])
            paths[-1].append(split[-1])

        root = {}
        for path in paths:
            branch = root.setdefault(path[0], [{}, []])
            for step in path[1:-1]:
                branch = branch[0].setdefault(step, [{}, []])
            branch[1].append(path[-1])
        self.deleter(root)
        return(root)

    def walker(self, courls: list):
        if isinstance(courls, list):
            for item in courls:
                yield item
        if isinstance(courls, dict):
            for item in courls.values():
                yield item

    def deleter(self, courls: list):
        for data in self.walker(courls):
            if data == [] or data == {}:
                courls.remove(data)
            self.deleter(data)

    def createTree(self, parent: str, files: dict | list):
        for key in files:
            if not parent:
                identifier = str(''.join(random.choices(string.ascii_uppercase + string.digits, k = 10)))
                self.tree.create_node(key, identifier)
            if not key:
                continue
            try:
                value = files[key]
                if parent:
                    identifier = str(''.join(random.choices(string.ascii_uppercase + string.digits, k = 10)))
                    self.tree.create_node(key, identifier, parent=parent)
                self.createTree(identifier, value)
            except:
                if type(key) == dict or type(key) == list:
                    self.createTree(parent, key)
                else:
                    identifier = str(''.join(random.choices(string.ascii_uppercase + string.digits, k = 10)))
                    self.tree.create_node(key, identifier, parent=parent)



class Keyspace:
    
    space = 0
    space_complete = 0
    
    
    @classmethod
    def increment_space(cls) -> None:
        cls.space += 1
        
    
    @classmethod
    def increment_space_by(cls, increment: int) -> None:
        cls.space += increment
        
        
    @classmethod
    def increment_space_complete(cls) -> None:
        cls.space_complete += 1
        
        
    @classmethod
    def increment_space_complete_by(cls, increment: int) -> None:
        cls.space_complete += increment


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
cachefile.py - file with classes handling cached TLDs (e.g. downloads, updates)

.. Licence MIT
.. codeauthor:: Jan Lipovský <janlipovsky@gmail.com>, janlipovsky.cz
.. contributors: https://github.com/lipoja/URLExtract/graphs/contributors
"""

import logging
import os
import tempfile
import urllib.request
from datetime import datetime
from urllib.error import URLError, HTTPError

import idna
import filelock
from appdirs import user_cache_dir


class CacheFileError(Exception):
    """
    Raised when some error occurred regarding file with cached TLDs.
    """
    pass


class CacheFile:
    """
    Class for working with cached TLDs in file.
    """

    # file name of cached list of TLDs downloaded from IANA
    _CACHE_FILE_NAME = 'tlds-alpha-by-domain.txt'
    _DATA_DIR = 'data'

    # name used in appdir
    _URLEXTRACT_NAME = "urlextract"

    def __init__(self, cache_dir=None):
        """
        :param str cache_dir: base path for TLD cache, defaults to data dir
        :raises: CacheFileError when cached file is not readable for user
        """

        self._logger = logging.getLogger(self._URLEXTRACT_NAME)

        self._user_defined_cache_dir = cache_dir
        self._default_cache_file = False

        # full path for cached file with list of TLDs
        self._tld_list_path = self._get_cache_file_path()
        if not os.access(self._tld_list_path, os.F_OK):
            self._logger.info(
                "Cache file not found in '%s'. "
                "Use URLExtract.update() to download newest version.",
                self._tld_list_path)
            self._logger.info(
                "Using default list of TLDs provided in urlextract package."
            )
            self._tld_list_path = self._get_default_cache_file_path()
            self._default_cache_file = True

    def _get_default_cache_dir(self):
        """
        Returns default cache directory (data directory)

        :raises: CacheFileError when default cached file does not is exist
        :return: path to default cache directory
        :rtype: str
        """

        return os.path.join(os.path.dirname(__file__), self._DATA_DIR)

    def _get_default_cache_file_path(self):
        """
        Returns default cache file path

        :return: default cache file path (to data directory)
        :rtype: str
        """

        default_list_path = os.path.join(
            self._get_default_cache_dir(), self._CACHE_FILE_NAME)

        if not os.access(default_list_path, os.F_OK):
            raise CacheFileError(
                "Default cache file does not exist "
                "'{}'!".format(default_list_path)
            )

        return default_list_path

    def _get_writable_cache_dir(self):
        """
        Get writable cache directory with fallback to user's cache directory
        and global temp directory

        :raises: CacheFileError when cached directory is not writable for user
        :return: path to cache directory
        :rtype: str
        """
        dir_path_data = self._get_default_cache_dir()

        if os.access(dir_path_data, os.W_OK):
            self._default_cache_file = True
            return dir_path_data

        dir_path_user = user_cache_dir(self._URLEXTRACT_NAME)
        if not os.path.exists(dir_path_user):
            try:
                os.makedirs(dir_path_user, exist_ok=True)
            except PermissionError:
                # if PermissionError exception is raised we should continue
                # and try to set the last fallback dir
                pass

        if os.access(dir_path_user, os.W_OK):
            return dir_path_user

        dir_path_temp = tempfile.gettempdir()
        if os.access(dir_path_temp, os.W_OK):
            return dir_path_temp

        raise CacheFileError("Cache directories are not writable.")

    def _get_cache_file_path(self):
        """
        Get path for cache file

        :raises: CacheFileError when cached directory is not writable for user
        :return: Full path to cached file with TLDs
        :rtype: str
        """
        if self._user_defined_cache_dir is None:
            # Tries to get writable cache dir with fallback to users data dir
            # and temp directory
            cache_dir = self._get_writable_cache_dir()
        else:
            cache_dir = self._user_defined_cache_dir
            if not os.access(self._user_defined_cache_dir, os.W_OK):
                raise CacheFileError(
                    "Cache directory {} is not writable.".format(
                        self._user_defined_cache_dir
                    )
                )

        # get path for cached file
        return os.path.join(cache_dir, self._CACHE_FILE_NAME)

    def _get_cache_lock_file_path(self):
        """
        Get path for cache file lock

        :raises: CacheFileError when cached directory is not writable for user
        :return: Full path to cached file lock
        :rtype: str
        """
        return self._get_cache_file_path()+'.lock'

    def _download_tlds_list(self):
        """
        Function downloads list of TLDs from IANA.
        LINK: https://data.iana.org/TLD/tlds-alpha-by-domain.txt

        :return: True if list was downloaded, False in case of an error
        :rtype: bool
        """
        url_list = 'https://data.iana.org/TLD/tlds-alpha-by-domain.txt'

        # Default cache file exist (set by _default_cache_file)
        # and we want to write permission
        if self._default_cache_file and \
                not os.access(self._tld_list_path, os.W_OK):
            self._logger.info("Default cache file is not writable.")
            self._tld_list_path = self._get_cache_file_path()
            self._logger.info(
                "Changed path of cache file to: %s",
                self._tld_list_path
            )

        if os.path.exists(self._tld_list_path) and \
                os.access(self._tld_list_path, os.F_OK) and \
                not os.access(self._tld_list_path, os.W_OK):
            self._logger.error("ERROR: Cache file is not writable for current "
                               "user. ({})".format(self._tld_list_path))
            return False

        req = urllib.request.Request(url_list)
        req.add_header('User-Agent', 'Mozilla/5.0 (Windows NT 6.0; '
                                     'WOW64; rv:24.0) Gecko/20100101 '
                                     'Firefox/24.0')
        try:
            with urllib.request.urlopen(req) as f:
                page = f.read().decode('utf-8')
        except HTTPError as e:
            self._logger.error("ERROR: Can not download list of TLDs. "
                               "(HTTPError: {})".format(e.reason))
            return False
        except URLError as e:
            self._logger.error("ERROR: Can not download list of TLDs. "
                               "(URLError: {})".format(e.reason))
            return False

        with filelock.FileLock(self._get_cache_lock_file_path()):
            with open(self._tld_list_path, 'w') as ftld:
                ftld.write(page)

        return True

    def _load_cached_tlds(self):
        """
        Loads TLDs from cached file to set.

        :return: Set of current TLDs
        :rtype: set
        """

        # check if cached file is readable
        if not os.access(self._tld_list_path, os.R_OK):
            self._logger.error("Cached file is not readable for current "
                               "user. ({})".format(self._tld_list_path))
            raise CacheFileError(
                "Cached file is not readable for current user."
            )

        set_of_tlds = set()

        with filelock.FileLock(self._get_cache_lock_file_path()):
            with open(self._tld_list_path, 'r') as f_cache_tld:
                for line in f_cache_tld:
                    tld = line.strip().lower()
                    # skip empty lines
                    if not tld:
                        continue
                    # skip comments
                    if tld[0] == '#':
                        continue

                    set_of_tlds.add("." + tld)
                    set_of_tlds.add("." + idna.decode(tld))

        return set_of_tlds

    def _get_last_cachefile_modification(self):
        """
        Get last modification of cache file with TLDs.

        :return: Date and time of last modification or
                 None when file does not exist
        :rtype: datetime|None
        """

        try:
            mtime = os.path.getmtime(self._tld_list_path)
        except OSError:
            return None

        return datetime.fromtimestamp(mtime)

def dns_cache_install():
    try:
        from dns_cache.resolver import ExceptionCachingResolver
        from dns import resolver as dnspython_resolver_module
        if not dnspython_resolver_module.default_resolver:
            dnspython_resolver_module.default_resolver = ExceptionCachingResolver()
        del dnspython_resolver_module
    except ImportError:
        pass

    try:
        from dns.resolver import LRUCache, Resolver, override_system_resolver, _resolver, default_resolver
    except ImportError:
        return

    if default_resolver:
        if not default_resolver.cache:
            default_resolver.cache = LRUCache()
        resolver = default_resolver
    elif _resolver and _resolver.cache:
        resolver = _resolver
    else:
        resolver = Resolver()
        resolver.cache = LRUCache()
    override_system_resolver(resolver)

dns_cache_install()    


import os

from ptlibs import tldparser
import urllib.parse


class Url:
    """
    A utility class for representing and manipulating URLs.

    This class provides helper methods for extracting paths, domains, and
    schemes from URLs, removing parameters, and ensuring standardized formats.
    It is designed to handle common URL transformations and parsing logic.

    Attributes:
        url (str): The original URL string.
    """

    def __init__(self, url: str) -> None:
        """
        Initialize a Url object.

        Args:
            url (str): The URL string to be stored and manipulated.
        """
        self.url = url

    def get_path_from_url(self, with_l_slash: bool = True, without_r_slash: bool = False) -> str:
        """
        Extract the path component from the URL.

        Args:
            with_l_slash (bool): If True, the returned path starts with '/'.
            without_r_slash (bool): If True and the URL is a directory (ends with '/'),
                the trailing slash will be removed.

        Returns:
            str: The path portion of the URL based on the given options.
        """
        url = self.get_url_without_parameters()
        out_r_slash = -1 if self.is_url_dictionary() and without_r_slash else None
        url = url.replace("//", "::")
        domain_len = url.find("/") if url.find("/")>0 else len(url)
        if with_l_slash:
            return url[domain_len:out_r_slash]
        else:
            return url[domain_len+1:out_r_slash]

    def get_url_without_parameters(self) -> str:
        """
        Return the URL without query parameters or fragments.

        Returns:
            str: The URL without '?' query parameters and '#' fragments.
        """
        return self.url.split("?")[0].split("#")[0]

    def is_url_dictionary(self) -> bool:
        """
        Check if the URL points to a directory.

        Returns:
            bool: True if the URL ends with '/', otherwise False.
        """
        return self.get_url_without_parameters().endswith("/")

    def standardize_url(self, domain_with_protocol: str) -> str:
        """
        Convert the stored URL to an absolute path form with a given domain.

        Args:
            domain_with_protocol (str): The full domain including the scheme
                (e.g., 'https://example.com').

        Returns:
            str: The standardized absolute URL.
        """
        path = self.url[len(domain_with_protocol):]
        if not path.startswith("/"):
            path = "/"
        abs = os.path.abspath(path)+"/" if path.endswith("/") and path !="/" else os.path.abspath(path)
        return domain_with_protocol + abs

    def get_domain_from_url(self, level=True, with_protocol=True) -> str:
        """
        Extract the domain from the URL.

        Args:
            level (bool): If True, return the full subdomain and domain.
                If False, return only the base domain.
            with_protocol (bool): If True, include the scheme (http/https).

        Returns:
            str: The extracted domain, optionally including subdomains and scheme.
        """
        extract = tldparser.parse(self.url)
        if extract.subdomain:
            extract.subdomain += "."
        if with_protocol:
            protocol = extract.scheme + "://" if extract.scheme else "http://"
        else:
            protocol = ""
        if level:
            return protocol + extract.subdomain + extract.domain + ("." if extract.suffix else "") + extract.suffix
        else:
            return protocol + extract.domain + ("." if extract.suffix else "") + extract.suffix

    def add_missing_scheme(self, scheme: str) -> str:
        """
        Ensure the URL has a scheme. If missing, prepend the given scheme.

        Args:
            scheme (str): The scheme to prepend (e.g., 'http', 'https').

        Returns:
            str: The URL with a scheme.
        """
        extract = urllib.parse.urlparse(self.url)
        if self.url and not (extract.scheme):
            return scheme + "://" + self.url
        else:
            return self.url


import os
import re
import requests

import ptlibs.ptprinthelper as ptprinthelper
from keyspace import Keyspace

from ptdataclasses.processedargumentoptions import ProcessedArgumentOptions
from utils.url import Url
from findings import FindingDetail, Findings

class ResponseProcessor:
    
    def __init__(self, domain_with_protocol: str, domain: str, args: ProcessedArgumentOptions) -> None:
        self.domain_with_protocol = domain_with_protocol
        self.domain = domain
        self.args = args
    
    
    def save_content(self, content: bytes, path: str, save_path: str) -> None:
        path = save_path + "/" + path
        dirname = os.path.dirname(path)
        if self.is_directory_traversal(path, save_path):
            return
        os.makedirs(dirname, exist_ok=True)
        if dirname + "/" != path:
            output_file = open(path,"wb")
            output_file.write(content)
            output_file.close()
            
        
    def is_directory_traversal(self, path: str, save_path: str) -> bool:
        current_directory = os.path.abspath(save_path)
        requested_path = os.path.abspath(path)
        common_prefix = os.path.commonprefix([requested_path, current_directory])
        return common_prefix != current_directory
    
        
    def check_content_type(self, response: requests.Response, request_url: str) -> tuple[str,str]:
        if response.url == request_url + "/" or Url(response.url).is_url_dictionary():
            if response.status_code == 200:
                return "directory", "[" + ptprinthelper.get_colored_text("D", "ERROR") + "] "
            else:
                return "directory", "[D] "
        else:
            return "file", "[F] "
        
        
    def get_response_history(self, history: list[requests.Response], json: bool, include_parameters: bool, urlpath: str, keyspace_for_directory: int) -> str:
        output = ""
        for resp in history:
            string = ptprinthelper.out_ifnot(f"[{resp.status_code}] [R]  {resp.url}  \u2794", "REDIR", json)
            output += ptprinthelper.add_spaces_to_eon(string) + "\n"
            self.parse_url_and_add_unigue_url_and_directories(resp.url, include_parameters, urlpath, keyspace_for_directory, resp)
            output += self.get_content_location(include_parameters, urlpath, keyspace_for_directory ,resp)
        return output
    
    
    def parse_url_and_add_unigue_url_and_directories(self, url: str, include_parameters: bool, urlpath: str, keyspace_for_directory: int, response: requests.Response = None) -> None:
        url_object = Url(url)
        if not include_parameters:
            url = url_object.get_url_without_parameters()
        path_from_url = url_object.get_path_from_url()
        segmented_path = [i for i in path_from_url.split("/")]
        last_segment_no = len(segmented_path)
        is_dir = url_object.is_url_dictionary()
        path = "/"
        for i, segment in enumerate(segmented_path):
            path += segment
            if (i != last_segment_no-1 or (i==last_segment_no-1 and is_dir)) and not self.args.url.endswith(path):
                self.add_unique_finding_to_findings(self.domain_with_protocol + path + "/", response if self.is_response(response) else None, urlpath)
                path += "/"
                path = re.sub('/{2,}', "/", path)
                self.add_unigue_directory_to_directories(path, urlpath, keyspace_for_directory)
                if i == last_segment_no-1 and self.is_response(response):
                    finding_detail = FindingDetail(url=self.domain_with_protocol + path, status_code=response.status_code, headers=response.headers)
                    Findings.details.append(finding_detail)
            else:
                self.add_unique_finding_to_findings(self.domain_with_protocol + path, response if self.is_response(response) else None, urlpath)
                path += "/"
                
    
    def add_unique_finding_to_findings(self, url: str, response: requests.Response, urlpath: str) -> None:
        url = url.replace("//", "/")
        url = url.replace(":/", "://")
        url = Url(url).standardize_url(self.domain_with_protocol)
        if self.args.is_star_in_domain:
            url = response.url
        if Url(url).is_url_dictionary() and url[:-1] in Findings.findings:
            Findings.findings.remove(url[:-1])
        if not url in Findings.findings and not url+"/" in Findings.findings and (Url(url).get_path_from_url().startswith(urlpath) or self.args.is_star_in_domain):
            Findings.findings.append(url)
            if self.is_response(response):
                findingDetail = FindingDetail(url=url, status_code=response.status_code, headers=response.headers)
                Findings.details.append(findingDetail)
            if self.args.backups and self.is_response(response) and not str(response.status_code).startswith("4"):
                Findings.findings2.append(url)
                Keyspace.increment_space()
                Keyspace.increment_space_complete()
            if self.args.parse:
                Keyspace.increment_space_complete()
                
    
    def is_response(self, response: requests.Response) -> bool:
        try:
            return True if response.status_code else False
        except:
            return False
        
        
    def add_unigue_directory_to_directories(self, directory: str, urlpath: str, keyspace_for_directory: int) -> None:
        directory = os.path.abspath(directory)
        directory = directory + "/" if directory != "/" else directory
        if self.args.recurse and directory.count('/') > self.args.max_depth+1:
            return
        if not directory in Findings.directories and self.args.recurse and directory.startswith(urlpath) and not self.started_path_with(directory, self.args.not_directories):
            Findings.directories.append(directory)
            Keyspace.increment_space_complete_by(keyspace_for_directory)
            
            
    def started_path_with(self, directory: str, not_directories: list[str]) -> bool:
        for nd in not_directories:
            if directory.startswith(nd):
                return True
        return False
    
    
    def get_content_location(self, include_parameters: bool, urlpath: str, keyspace_for_directory: int, response: requests.Response) -> str:
        output = ""
        try:
            if response.headers['Content-Location']:
                content_location = self.get_string_before_last_char(response.url, "/") + "/" +response.headers['Content-Location']
                string = ptprinthelper.out_ifnot(f"[-->] [L]  {content_location}", "OK", self.args.json)
            output += ptprinthelper.add_spaces_to_eon(string)
            self.parse_url_and_add_unigue_url_and_directories(content_location, include_parameters, urlpath, keyspace_for_directory, response)
        except:
            pass
        return output
    
    
    def get_string_before_last_char(self, string: str, chars: list[str]) -> str:
        for char in chars:
            string = string.rpartition(char)[0]
        return string
    
    
    def parse_html_find_and_add_urls(self, response: requests.Response, include_parameters: bool, urlpath: str, keyspace_for_directory: int, domain_protocol: str) -> str:
        if self.args.parse:
            output = "\n"
            urls = self.find_urls_in_html(response, domain_protocol)
            for url in urls:
                string = ptprinthelper.out_ifnot(f"           {url}", "PARSED", self.args.json)
                output += ptprinthelper.add_spaces_to_eon(string) + "\n"
                self.parse_url_and_add_unigue_url_and_directories(url, include_parameters, urlpath, keyspace_for_directory)
            return output.rstrip()
        else:
            return ""
        
        
    def find_urls_in_html(self, response: requests.Response, domain_protocol: str) -> list[str]:
        page_content = response.text
        if self.args.include_parameters:
            absolute_url_patern = re.compile(r'(https?)(:\/\/' + self.domain.replace(".", "\\.") + ')(\/?[^\[\]\'"><\s]*)?[\'"><\s]', flags=re.IGNORECASE)
            relative_url_patern = re.compile(r'(sitemap: |allow: | href=[\'"]*| src=[\'"]*)([^\[\]\\\'"\s<>]*)', flags=re.IGNORECASE)
        else:
            absolute_url_patern = re.compile(r'(https?)(:\/\/' + self.domain.replace(".", "\\.") + ')(\/[^\[\]\'"><?#\s]*)?[\'">?#<\s]', flags=re.IGNORECASE)
            relative_url_patern = re.compile(r'(sitemap: |allow: | href=[\'"]*| src=[\'"]*)([^\[\]\\\'"?#\s<>]*)', flags=re.IGNORECASE)
        max_parsed_content_length = self.get_last_parsed_character_index(response)
        all_urls = list({''.join(result) for result in absolute_url_patern.findall(page_content, endpos=max_parsed_content_length)})
        all_rel_urls = list({result[1] for result in relative_url_patern.findall(page_content, endpos=max_parsed_content_length)})
        for url in all_rel_urls:
            if not url in ["", "/", "?", "#"] and not url.lower().startswith(("mailto:", "tel:", "news:", "ftp:", "ftps:", "data:", "javascript:", "vbscript:")):
                absurl = self.rel2abs(response.url, url, domain_protocol)
                if absurl:
                    all_urls.append(absurl)
        return list(dict.fromkeys(list(all_urls)))
    
    
    def get_last_parsed_character_index(self, response: requests.Response) -> int:
        last_parsed_character = self.args.content_length / self.get_encoding_bytes_per_char(response.encoding)
        return int(last_parsed_character)
    
    
    def get_encoding_bytes_per_char(self, encoding: str) -> int:
        char = "A"
        encoded_bytes = char.encode(encoding)
        return len(encoded_bytes)
    
    
    def rel2abs(self, location: str, url: str, domain_protocol: str) -> str:
        if re.match('^\w{2,5}:\/\/', url):
            if re.match('^\w{2,5}:\/\/' + self.domain.replace(".", "\\."), url):
                return url
            else:
                return None
        elif url.startswith("//"):
            if url.startswith("//" + self.domain):
                return domain_protocol + url
            else:
                return None
        elif url.startswith("/"):
            return self.domain_with_protocol + url
        else: 
            if url.startswith("?") or url.startswith("#"):
                return self.get_string_before_last_char(location, ["?", "#"]) + url
            else:
                return self.get_string_before_last_char(location, ["/"]) + "/" + url
            
            
    def get_content_type_and_length(self, headers: dict[str,str]) -> tuple[str, str]:
            try:
                c_l = headers['content-length']
            except:
                c_l = "?"
            try:
                c_t = headers['Content-Type'].split(";")[0]
            except:
                c_t = "unknown"
            return c_t, c_l
        
        
    def add_unigue_technology_to_technologies(self, technology: str) -> None:
        if not technology in Findings.technologies:
            Findings.technologies.append(technology)
            
    
    def content_shorter_than_maximum(self, response: requests.Response) -> bool:
        _, c_l = self.get_content_type_and_length(response.headers)
        if not c_l.isdigit():
            return False
        
        content_length = int(c_l)
        return content_length < self.args.content_length


from ptlibs.threads.arraylock import ThreadSafeArray

from ptdataclasses.findingdetail import FindingDetail


class Findings:
    findings = ThreadSafeArray[str]()
    findings2 = ThreadSafeArray[str]()
    details = ThreadSafeArray[FindingDetail]()
    directories = ThreadSafeArray[str]()
    technologies = ThreadSafeArray[str]()
    visited = ThreadSafeArray[str]()


import setuptools
from ptwebdiscover._version import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ptwebdiscover",
    description="Web Source Discovery Tool",
    url="https://www.penterep.com/",
    author="Penterep",
    author_email="info@penterep.com",
    version=__version__,
    license="GPLv3",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: Implementation :: CPython",
        "Environment :: Console",
        "Topic :: Security",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)"
    ],
    python_requires = '>=3.9',
    install_requires=["ptlibs>=1.0.7,<2", "appdirs", "treelib", "filelock"],
    entry_points = {'console_scripts': ['ptwebdiscover = ptwebdiscover.ptwebdiscover:main']},
    include_package_data= True,
    long_description=long_description,
    long_description_content_type="text/markdown",
)


import re

class RobotsTxtParser:
    def __init__(self, robots_txt: str):
        # Stores the original paths as strings
        self.allows = set()
        self.disallows = set()
        # Stores compiled regex for internal matching
        self._regex_allow = []
        self._regex_disallow = []
        self._parse(robots_txt)

    def _wildcard_to_regex(self, pattern: str) -> re.Pattern:
        """
        Convert robots.txt pattern with '*' and '$' to regex.
        '*' -> '.*', '$' -> end-of-string
        """
        regex = re.escape(pattern)
        regex = regex.replace(r"\*", ".*")
        if regex.endswith(r"\$"):
            regex = regex[:-2] + "$"
        else:
            regex += ".*"
        return re.compile("^" + regex)

    def _parse(self, robots_txt: str):
        for line in robots_txt.splitlines():
            line = line.strip()
            if not line or line.startswith("#"):
                continue

            if ":" not in line:
                continue

            directive, value = line.split(":", 1)
            directive = directive.strip().lower()
            value = value.strip()
            if not value:
                continue

            # Store string in sets
            if directive == "allow":
                self.allows.add(value)
                self._regex_allow.append(self._wildcard_to_regex(value))
            elif directive == "disallow":
                self.disallows.add(value)
                self._regex_disallow.append(self._wildcard_to_regex(value))

    def is_allowed(self, url_path: str) -> bool:
        """
        Returns True if the path is allowed, False if disallowed.
        Follows longest match precedence.
        """
        matched_rule = None
        matched_length = -1
        allowed = True  # default allow

        for rule_type, patterns in [("allow", self._regex_allow), ("disallow", self._regex_disallow)]:
            for regex in patterns:
                match = regex.match(url_path)
                if match:
                    length = len(match.group(0))
                    if length > matched_length:
                        matched_rule = rule_type
                        matched_length = length
                        allowed = (rule_type == "allow")

        return allowed

    def get_allows(self):
        return self.allows

    def get_disallows(self):
        return self.disallows
